@@ -78,8 +78,12 @@ static void  swap_header_copy(HASHHDR *, HASHHDR *);
 #endif
 
 /* Macros for min/max.  */
+#ifndef MIN
 #define MIN(a,b) (((a)<(b))?(a):(b))
+#endif
+#ifndef MAX
 #define MAX(a,b) (((a)>(b))?(a):(b))
+#endif
 
 /* Fast arithmetic, relying on powers of 2, */
 #define MOD(x, y)		((x) & ((y) - 1))
@@ -105,6 +109,7 @@ __hash_open(file, flags, mode, info, dflags)
 	const HASHINFO *info;	/* Special directives for create */
 {
 	HTAB *hashp;
+
 	struct stat statbuf;
 	DB *dbp;
 	int bpages, hdrsize, new_table, nsegs, save_errno;
@@ -128,7 +133,11 @@ __hash_open(file, flags, mode, info, dflags)
 
 	new_table = 0;
 	if (!file || (flags & O_TRUNC) ||
+#ifdef __USE_INTERNAL_STAT64
+	    (stat64(file, &statbuf) && (errno == ENOENT))) {
+#else
 	    (stat(file, &statbuf) && (errno == ENOENT))) {
+#endif
 		if (errno == ENOENT)
 			errno = 0; /* Just in case someone looks at errno */
 		new_table = 1;
@@ -140,7 +149,11 @@ __hash_open(file, flags, mode, info, dflags)
 		/* if the .db file is empty, and we had permission to create
 		   a new .db file, then reinitialize the database */
 		if ((flags & O_CREAT) &&
+#ifdef __USE_INTERNAL_STAT64
+		     fstat64(hashp->fp, &statbuf) == 0 && statbuf.st_size == 0)
+#else
 		     fstat(hashp->fp, &statbuf) == 0 && statbuf.st_size == 0)
+#endif
 			new_table = 1;
 
 #ifdef HAVE_FCNTL
@@ -314,7 +327,11 @@ init_hash(hashp, file, info)
 
 	/* Fix bucket size to be optimal for file system */
 	if (file != NULL) {
+#ifdef __USE_INTERNAL_STAT64
+		if (stat64(file, &statbuf))
+#else
 		if (stat(file, &statbuf))
+#endif
 			return (NULL);
 		hashp->BSIZE = statbuf.st_blksize;
 		hashp->BSHIFT = __log2(hashp->BSIZE);
