@@ -1,6 +1,4 @@
-/* No user fns here. Pesch 15apr92. */
-
-/*
+/*-
  * Copyright (c) 1990 The Regents of the University of California.
  * All rights reserved.
  *
@@ -17,11 +15,103 @@
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  */
 
+/*
+FUNCTION
+<<vscanf>>, <<vfscanf>>, <<vsscanf>>---format argument list
+
+INDEX
+	vscanf
+INDEX
+	vfscanf
+INDEX
+	vsscanf
+
+ANSI_SYNOPSIS
+	#include <stdio.h>
+	#include <stdarg.h>
+	int vscanf(const char *<[fmt]>, va_list <[list]>);
+	int vfscanf(FILE *<[fp]>, const char *<[fmt]>, va_list <[list]>);
+	int vsscanf(const char *<[str]>, const char *<[fmt]>, va_list <[list]>);
+
+	int _vscanf_r(struct _reent *<[reent]>, const char *<[fmt]>, 
+                       va_list <[list]>);
+	int _vfscanf_r(struct _reent *<[reent]>, FILE *<[fp]>, const char *<[fmt]>, 
+                       va_list <[list]>);
+	int _vsscanf_r(struct _reent *<[reent]>, const char *<[str]>, const char *<[fmt]>, 
+                       va_list <[list]>);
+
+TRAD_SYNOPSIS
+	#include <stdio.h>
+	#include <varargs.h>
+	int vscanf( <[fmt]>, <[ist]>)
+	char *<[fmt]>;
+	va_list <[list]>;
+
+	int vfscanf( <[fp]>, <[fmt]>, <[list]>)
+	FILE *<[fp]>;
+	char *<[fmt]>;
+	va_list <[list]>;
+	
+	int vsscanf( <[str]>, <[fmt]>, <[list]>)
+	char *<[str]>;
+	char *<[fmt]>;
+	va_list <[list]>;
+
+	int _vscanf_r( <[reent]>, <[fmt]>, <[ist]>)
+	struct _reent *<[reent]>;
+	char *<[fmt]>;
+	va_list <[list]>;
+
+	int _vfscanf_r( <[reent]>, <[fp]>, <[fmt]>, <[list]>)
+	struct _reent *<[reent]>;
+	FILE *<[fp]>;
+	char *<[fmt]>;
+	va_list <[list]>;
+	
+	int _vsscanf_r( <[reent]>, <[str]>, <[fmt]>, <[list]>)
+	struct _reent *<[reent]>;
+	char *<[str]>;
+	char *<[fmt]>;
+	va_list <[list]>;
+
+DESCRIPTION
+<<vscanf>>, <<vfscanf>>, and <<vsscanf>> are (respectively) variants
+of <<scanf>>, <<fscanf>>, and <<sscanf>>.  They differ only in 
+allowing their caller to pass the variable argument list as a 
+<<va_list>> object (initialized by <<va_start>>) rather than 
+directly accepting a variable number of arguments.
+
+RETURNS
+The return values are consistent with the corresponding functions:
+<<vscanf>> returns the number of input fields successfully scanned,
+converted, and stored; the return value does not include scanned
+fields which were not stored.  
+
+If <<vscanf>> attempts to read at end-of-file, the return value 
+is <<EOF>>.
+
+If no fields were stored, the return value is <<0>>.
+
+The routines <<_vscanf_r>>, <<_vfscanf_f>>, and <<_vsscanf_r>> are
+reentrant versions which take an additional first parameter which points to the
+reentrancy structure.
+
+PORTABILITY
+These are GNU extensions.
+
+Supporting OS subroutines required:
+*/
+
 #include <_ansi.h>
+#include <reent.h>
+#include <newlib.h>
 #include <ctype.h>
+#include <wctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <limits.h>
+#include <wchar.h>
+#include <string.h>
 #ifdef _HAVE_STDC
 #include <stdarg.h>
 #else
@@ -29,9 +119,20 @@
 #endif
 #include "local.h"
 
-#ifndef	NO_FLOATING_POINT
+#ifdef INTEGER_ONLY
+#define VFSCANF vfiscanf
+#define _VFSCANF_R _vfiscanf_r
+#define __SVFSCANF __svfiscanf
+#define __SVFSCANF_R __svfiscanf_r
+#else
+#define VFSCANF vfscanf
+#define _VFSCANF_R _vfscanf_r
+#define __SVFSCANF __svfscanf
+#define __SVFSCANF_R __svfscanf_r
+#ifndef NO_FLOATING_POINT
 #define FLOATING_POINT
 #endif
+#endif
 
 #ifdef FLOATING_POINT
 #include <float.h>
@@ -40,13 +141,19 @@
    This could be changed in the future should the _ldtoa_r code be
    preferred over _dtoa_r.  */
 #define _NO_LONGDBL
-#if defined WANT_IO_LONG_DBL && (LDBL_MANT_DIG > DBL_MANT_DIG)
+#if defined _WANT_IO_LONG_DOUBLE && (LDBL_MANT_DIG > DBL_MANT_DIG)
 #undef _NO_LONGDBL
 extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 #endif
 
 #include "floatio.h"
-#define	BUF	(MAXEXP+MAXFRACT+3)	/* 3 = sign + decimal point + NUL */
+
+#if ((MAXEXP+MAXFRACT+3) > MB_LEN_MAX)
+#  define BUF (MAXEXP+MAXFRACT+3)        /* 3 = sign + decimal point + NUL */
+#else
+#  define BUF MB_LEN_MAX
+#endif
+
 /* An upper bound for how long a long prints in decimal.  4 / 13 approximates
    log (2).  Add one char for roundoff compensation and one for the sign.  */
 #define MAX_LONG_LEN ((CHAR_BIT * sizeof (long)  - 1) * 4 / 13 + 2)
@@ -54,16 +161,22 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 #define	BUF	40
 #endif
 
+#define _NO_LONGLONG
+#if defined _WANT_IO_LONG_LONG && defined __GNUC__
+# undef _NO_LONGLONG
+#endif
+
 /*
  * Flags used during conversion.
  */
 
 #define	LONG		0x01	/* l: long or double */
-#define	LONGDBL		0x02	/* L: long double */
+#define	LONGDBL		0x02	/* L/ll: long double or long long */
 #define	SHORT		0x04	/* h: short */
-#define	SUPPRESS	0x08	/* suppress assignment */
-#define	POINTER		0x10	/* weird %p pointer (`fake hex') */
-#define	NOSKIP		0x20	/* do not skip blanks */
+#define CHAR		0x08	/* hh: 8 bit integer */
+#define	SUPPRESS	0x10	/* suppress assignment */
+#define	POINTER		0x20	/* weird %p pointer (`fake hex') */
+#define	NOSKIP		0x40	/* do not skip blanks */
 
 /*
  * The following are used in numeric conversions only:
@@ -71,14 +184,15 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
  * SIGNOK, NDIGITS, PFXOK, and NZDIGITS are for integral.
  */
 
-#define	SIGNOK		0x40	/* +/- is (still) legal */
-#define	NDIGITS		0x80	/* no digits detected */
+#define	SIGNOK		0x80	/* +/- is (still) legal */
+#define	NDIGITS		0x100	/* no digits detected */
 
-#define	DPTOK		0x100	/* (float) decimal point is still legal */
-#define	EXPOK		0x200	/* (float) exponent (e+3, etc) still legal */
+#define	DPTOK		0x200	/* (float) decimal point is still legal */
+#define	EXPOK		0x400	/* (float) exponent (e+3, etc) still legal */
 
-#define	PFXOK		0x100	/* 0x prefix is (still) legal */
-#define	NZDIGITS	0x200	/* no zero digits detected */
+#define	PFXOK		0x200	/* 0x prefix is (still) legal */
+#define	NZDIGITS	0x400	/* no zero digits detected */
+#define	NNZDIGITS	0x800	/* no non-zero digits detected */
 
 /*
  * Conversion types.
@@ -96,6 +210,10 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 #define u_char char
 #define u_long unsigned long
 
+#ifndef _NO_LONGLONG
+typedef unsigned long long u_long_long;
+#endif
+
 /*static*/ u_char *__sccl ();
 
 /*
@@ -104,11 +222,46 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 
 #define BufferEmpty (fp->_r <= 0 && __srefill(fp))
 
+#ifndef _REENT_ONLY
+
 int
-__svfscanf (fp, fmt0, ap)
-     register FILE *fp;
-     char _CONST *fmt0;
-     va_list ap;
+_DEFUN(VFSCANF, (fp, fmt, ap), 
+       register FILE *fp _AND 
+       _CONST char *fmt _AND 
+       va_list ap)
+{
+  CHECK_INIT(fp);
+  return __SVFSCANF_R (_REENT, fp, fmt, ap);
+}
+
+int
+_DEFUN(__SVFSCANF, (fp, fmt0, ap),
+       register FILE *fp _AND
+       char _CONST *fmt0 _AND
+       va_list ap)
+{
+  return __SVFSCANF_R (_REENT, fp, fmt0, ap);
+}
+
+#endif /* !_REENT_ONLY */
+
+int
+_DEFUN(_VFSCANF_R, (data, fp, fmt, ap),
+       struct _reent *data _AND 
+       register FILE *fp   _AND 
+       _CONST char *fmt    _AND 
+       va_list ap)
+{
+  return __SVFSCANF_R (data, fp, fmt, ap);
+}
+
+
+int
+_DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
+       struct _reent *rptr _AND
+       register FILE *fp   _AND
+       char _CONST *fmt0   _AND
+       va_list ap)
 {
   register u_char *fmt = (u_char *) fmt0;
   register int c;		/* character from format, or conversion */
@@ -122,45 +275,50 @@ __svfscanf (fp, fmt0, ap)
   int base = 0;			/* base argument to strtol/strtoul */
   int nbytes = 1;               /* number of bytes read from fmt string */
   wchar_t wc;                   /* wchar to use to read format string */
+  wchar_t *wcp;                 /* handy wide character pointer */
+  size_t mbslen;                /* length of converted multibyte sequence */
+  mbstate_t state;              /* value to keep track of multibyte state */
 
   u_long (*ccfn) () = 0;	/* conversion function (strtol/strtoul) */
   char ccltab[256];		/* character class table for %[...] */
   char buf[BUF];		/* buffer for numeric conversions */
   char *lptr;                   /* literal pointer */
-#ifdef MB_CAPABLE
-  int state = 0;                /* value to keep track of multibyte state */
-#endif
 
+  char *cp;
   short *sp;
   int *ip;
   float *flp;
   _LONG_DOUBLE *ldp;
   double *dp;
   long *lp;
+#ifndef _NO_LONGLONG
+  long long *llp;
+#endif
 
   /* `basefix' is used to avoid `if' tests in the integer scanner */
   static _CONST short basefix[17] =
     {10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
 
+  _flockfile (fp);
+ 
   nassigned = 0;
   nread = 0;
   for (;;)
     {
-#ifndef MB_CAPABLE
+#ifndef _MB_CAPABLE
       wc = *fmt;
 #else
-      nbytes = _mbtowc_r (_REENT, &wc, fmt, MB_CUR_MAX, &state);
+      memset (&state, '\0', sizeof (state));
+      nbytes = _mbtowc_r (rptr, &wc, fmt, MB_CUR_MAX, &state);
 #endif
       fmt += nbytes;
       if (wc == 0)
-	return nassigned;
+	goto all_done;
       if (nbytes == 1 && isspace (wc))
 	{
 	  for (;;)
 	    {
-	      if (BufferEmpty)
-		return nassigned;
-	      if (!isspace (*fp->_p))
+	      if (BufferEmpty || !isspace (*fp->_p))
 		break;
 	      nread++, fp->_r--, fp->_p++;
 	    }
@@ -200,13 +358,25 @@ __svfscanf (fp, fmt0, ap)
 	  flags |= SUPPRESS;
 	  goto again;
 	case 'l':
-	  flags |= LONG;
+	  if (*fmt == 'l')	/* Check for 'll' = long long (SUSv3) */
+	    {
+	      ++fmt;
+	      flags |= LONGDBL;
+	    }
+	  else
+	    flags |= LONG;
 	  goto again;
 	case 'L':
 	  flags |= LONGDBL;
 	  goto again;
 	case 'h':
-	  flags |= SHORT;
+	  if (*fmt == 'h')	/* Check for 'hh' = char int (SUSv3) */
+	    {
+	      ++fmt;
+	      flags |= CHAR;
+	    }
+	  else
+	    flags |= SHORT;
 	  goto again;
 
 	case '0':
@@ -235,13 +405,13 @@ __svfscanf (fp, fmt0, ap)
 	  /* FALLTHROUGH */
 	case 'd':
 	  c = CT_INT;
-	  ccfn = (u_long (*)())strtol;
+	  ccfn = (u_long (*)())_strtol_r;
 	  base = 10;
 	  break;
 
 	case 'i':
 	  c = CT_INT;
-	  ccfn = (u_long (*)())strtol;
+	  ccfn = (u_long (*)())_strtol_r;
 	  base = 0;
 	  break;
 
@@ -250,13 +420,13 @@ __svfscanf (fp, fmt0, ap)
 	  /* FALLTHROUGH */
 	case 'o':
 	  c = CT_INT;
-	  ccfn = strtoul;
+	  ccfn = _strtoul_r;
 	  base = 8;
 	  break;
 
 	case 'u':
 	  c = CT_INT;
-	  ccfn = strtoul;
+	  ccfn = _strtoul_r;
 	  base = 10;
 	  break;
 
@@ -264,7 +434,7 @@ __svfscanf (fp, fmt0, ap)
 	case 'x':
 	  flags |= PFXOK;	/* enable 0x prefixing */
 	  c = CT_INT;
-	  ccfn = strtoul;
+	  ccfn = _strtoul_r;
 	  base = 16;
 	  break;
 
@@ -279,6 +449,9 @@ __svfscanf (fp, fmt0, ap)
 	  c = CT_FLOAT;
 	  break;
 #endif
+        case 'S':
+          flags |= LONG;
+          /* FALLTHROUGH */
 
 	case 's':
 	  c = CT_STRING;
@@ -290,6 +463,10 @@ __svfscanf (fp, fmt0, ap)
 	  c = CT_CCL;
 	  break;
 
+        case 'C':
+          flags |= LONG;
+          /* FALLTHROUGH */
+
 	case 'c':
 	  flags |= NOSKIP;
 	  c = CT_CHAR;
@@ -298,14 +475,19 @@ __svfscanf (fp, fmt0, ap)
 	case 'p':		/* pointer format is like hex */
 	  flags |= POINTER | PFXOK;
 	  c = CT_INT;
-	  ccfn = strtoul;
+	  ccfn = _strtoul_r;
 	  base = 16;
 	  break;
 
 	case 'n':
 	  if (flags & SUPPRESS)	/* ??? */
 	    continue;
-	  if (flags & SHORT)
+	  if (flags & CHAR)
+	    {
+	      cp = va_arg (ap, char *);
+	      *cp = nread;
+	    }
+	  else if (flags & SHORT)
 	    {
 	      sp = va_arg (ap, short *);
 	      *sp = nread;
@@ -315,6 +497,13 @@ __svfscanf (fp, fmt0, ap)
 	      lp = va_arg (ap, long *);
 	      *lp = nread;
 	    }
+#ifndef _NO_LONGLONG
+	  else if (flags & LONGDBL)
+	    {
+	      llp = va_arg (ap, long long*);
+	      *llp = nread;
+	    }
+#endif
 	  else
 	    {
 	      ip = va_arg (ap, int *);
@@ -326,13 +515,14 @@ __svfscanf (fp, fmt0, ap)
 	   * Disgusting backwards compatibility hacks.	XXX
 	   */
 	case '\0':		/* compat */
+	  _funlockfile (fp);
 	  return EOF;
 
 	default:		/* compat */
 	  if (isupper (c))
 	    flags |= LONG;
 	  c = CT_INT;
-	  ccfn = (u_long (*)())strtol;
+	  ccfn = (u_long (*)())_strtol_r;
 	  base = 10;
 	  break;
 	}
@@ -355,9 +545,7 @@ __svfscanf (fp, fmt0, ap)
 	      if (--fp->_r > 0)
 		fp->_p++;
 	      else
-#ifndef CYGNUS_NEC
 	      if (__srefill (fp))
-#endif
 		goto input_failure;
 	    }
 	  /*
@@ -377,10 +565,47 @@ __svfscanf (fp, fmt0, ap)
 	  /* scan arbitrary characters (sets NOSKIP) */
 	  if (width == 0)
 	    width = 1;
-	  if (flags & SUPPRESS)
+          if (flags & LONG) 
+            {
+              if ((flags & SUPPRESS) == 0)
+                wcp = va_arg (ap, wchar_t *);
+              else
+                wcp = NULL;
+              n = 0;
+              while (width != 0) 
+                {
+                  if (n == MB_CUR_MAX)
+                    goto input_failure;
+                  buf[n++] = *fp->_p;
+                  fp->_r -= 1;
+                  fp->_p += 1;
+                  memset ((_PTR)&state, '\0', sizeof (mbstate_t));
+                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state)) 
+                                                         == (size_t)-1)
+                    goto input_failure; /* Invalid sequence */
+                  if (mbslen == 0 && !(flags & SUPPRESS))
+                    *wcp = L'\0';
+                  if (mbslen != (size_t)-2) /* Incomplete sequence */
+                    {
+                      nread += n;
+                      width -= 1;
+                      if (!(flags & SUPPRESS))
+                        wcp += 1;
+                      n = 0;
+                    }
+                  if (BufferEmpty) 
+	            {
+                      if (n != 0) 
+                        goto input_failure;
+                      break;
+                    }
+                }
+              if (!(flags & SUPPRESS))
+                nassigned++;
+            } 
+          else if (flags & SUPPRESS) 
 	    {
 	      size_t sum = 0;
-
 	      for (;;)
 		{
 		  if ((n = fp->_r) < (int)width)
@@ -388,16 +613,12 @@ __svfscanf (fp, fmt0, ap)
 		      sum += n;
 		      width -= n;
 		      fp->_p += n;
-#ifndef CYGNUS_NEC
 		      if (__srefill (fp))
 			{
-#endif
 			  if (sum == 0)
 			    goto input_failure;
 			  break;
-#ifndef CYGNUS_NEC
 			}
-#endif
 		    }
 		  else
 		    {
@@ -411,27 +632,11 @@ __svfscanf (fp, fmt0, ap)
 	    }
 	  else
 	    {
-#ifdef CYGNUS_NEC
-	      /* Kludge city for the moment */
-	      char *dest = va_arg (ap, char *);
-	      int n = width;
-	      if (fp->_r == 0)
-		goto input_failure;
-
-	      while (n && fp->_r)
-		{
-		  *dest++ = *(fp->_p++);
-		  n--;
-		  fp->_r--;
-		  nread++;
-		}
-#else
 	      size_t r = fread ((_PTR) va_arg (ap, char *), 1, width, fp);
 
 	      if (r == 0)
 		goto input_failure;
 	      nread += r;
-#endif
 	      nassigned++;
 	    }
 	  break;
@@ -487,8 +692,56 @@ __svfscanf (fp, fmt0, ap)
 	case CT_STRING:
 	  /* like CCL, but zero-length string OK, & no NOSKIP */
 	  if (width == 0)
-	    width = ~0;
-	  if (flags & SUPPRESS)
+            width = (size_t)~0;
+          if (flags & LONG) 
+            {
+              /* Process %S and %ls placeholders */
+              if ((flags & SUPPRESS) == 0)
+                wcp = va_arg (ap, wchar_t *);
+              else
+                wcp = &wc;
+              n = 0;
+              while (!isspace (*fp->_p) && width != 0) 
+                {
+                  if (n == MB_CUR_MAX)
+                    goto input_failure;
+                  buf[n++] = *fp->_p;
+                  fp->_r -= 1;
+                  fp->_p += 1;
+                  memset ((_PTR)&state, '\0', sizeof (mbstate_t));
+                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state)) 
+                                                        == (size_t)-1)
+                    goto input_failure;
+                  if (mbslen == 0)
+                    *wcp = L'\0';
+                  if (mbslen != (size_t)-2) /* Incomplete sequence */
+                    {
+                      if (iswspace(*wcp)) 
+                        {
+                          while (n != 0)
+                            ungetc (buf[--n], fp);
+                          break;
+                        }
+                      nread += n;
+                      width -= 1;
+                      if ((flags & SUPPRESS) == 0)
+                        wcp += 1;
+                      n = 0;
+                    }
+                  if (BufferEmpty) 
+                    {
+                      if (n != 0)
+                        goto input_failure;
+                      break;
+                    }
+                }
+              if (!(flags & SUPPRESS)) 
+                {
+                  *wcp = L'\0';
+                  nassigned++;
+                }
+            }
+          else if (flags & SUPPRESS) 
 	    {
 	      n = 0;
 	      while (!isspace (*fp->_p))
@@ -520,17 +773,21 @@ __svfscanf (fp, fmt0, ap)
 	  continue;
 
 	case CT_INT:
+	{
 	  /* scan an integer as if by strtol/strtoul */
+	  unsigned width_left = 0;
+	  int skips = 0;
 #ifdef hardway
 	  if (width == 0 || width > sizeof (buf) - 1)
-	    width = sizeof (buf) - 1;
 #else
 	  /* size_t is unsigned, hence this optimisation */
-	  if (--width > sizeof (buf) - 2)
-	    width = sizeof (buf) - 2;
-	  width++;
+	  if (width - 1 > sizeof (buf) - 2)
 #endif
-	  flags |= SIGNOK | NDIGITS | NZDIGITS;
+	    {
+	      width_left = width - (sizeof (buf) - 1);
+	      width = sizeof (buf) - 1;
+	    }
+	  flags |= SIGNOK | NDIGITS | NZDIGITS | NNZDIGITS;
 	  for (p = buf; width; width--)
 	    {
 	      c = *fp->_p;
@@ -550,16 +807,26 @@ __svfscanf (fp, fmt0, ap)
 		   * will turn it off if we have scanned any nonzero digits).
 		   */
 		case '0':
+		  if (! (flags & NNZDIGITS))
+		    goto ok;
 		  if (base == 0)
 		    {
 		      base = 8;
 		      flags |= PFXOK;
 		    }
 		  if (flags & NZDIGITS)
-		    flags &= ~(SIGNOK | NZDIGITS | NDIGITS);
-		  else
-		    flags &= ~(SIGNOK | PFXOK | NDIGITS);
-		  goto ok;
+		    {
+		      flags &= ~(SIGNOK | NZDIGITS | NDIGITS);
+		      goto ok;
+		    }
+		  flags &= ~(SIGNOK | PFXOK | NDIGITS);
+		  if (width_left)
+		    {
+		      width_left--;
+		      width++;
+		    }
+		  ++skips;
+		  goto skip;
 
 		  /* 1 through 7 always legal */
 		case '1':
@@ -570,7 +837,7 @@ __svfscanf (fp, fmt0, ap)
 		case '6':
 		case '7':
 		  base = basefix[base];
-		  flags &= ~(SIGNOK | PFXOK | NDIGITS);
+		  flags &= ~(SIGNOK | PFXOK | NDIGITS | NNZDIGITS);
 		  goto ok;
 
 		  /* digits 8 and 9 ok iff decimal or hex */
@@ -579,7 +846,7 @@ __svfscanf (fp, fmt0, ap)
 		  base = basefix[base];
 		  if (base <= 8)
 		    break;	/* not legal here */
-		  flags &= ~(SIGNOK | PFXOK | NDIGITS);
+		  flags &= ~(SIGNOK | PFXOK | NDIGITS | NNZDIGITS);
 		  goto ok;
 
 		  /* letters ok iff hex */
@@ -598,7 +865,7 @@ __svfscanf (fp, fmt0, ap)
 		  /* no need to fix base here */
 		  if (base <= 10)
 		    break;	/* not legal here */
-		  flags &= ~(SIGNOK | PFXOK | NDIGITS);
+		  flags &= ~(SIGNOK | PFXOK | NDIGITS | NNZDIGITS);
 		  goto ok;
 
 		  /* sign ok only as first character */
@@ -618,6 +885,10 @@ __svfscanf (fp, fmt0, ap)
 		    {
 		      base = 16;/* if %i */
 		      flags &= ~PFXOK;
+		      /* We must reset the NZDIGITS and NDIGITS
+		         flags that would have been unset by seeing
+			 the zero that preceded the X or x.  */
+		      flags |= NZDIGITS | NDIGITS;
 		      goto ok;
 		    }
 		  break;
@@ -633,12 +904,11 @@ __svfscanf (fp, fmt0, ap)
 	       * c is legal: store it and look at the next.
 	       */
 	      *p++ = c;
+	    skip:
 	      if (--fp->_r > 0)
 		fp->_p++;
 	      else
-#ifndef CYGNUS_NEC
 	      if (__srefill (fp))
-#endif
 		break;		/* EOF */
 	    }
 	  /*
@@ -663,9 +933,14 @@ __svfscanf (fp, fmt0, ap)
 	      u_long res;
 
 	      *p = 0;
-	      res = (*ccfn) (buf, (char **) NULL, base);
+	      res = (*ccfn) (rptr, buf, (char **) NULL, base);
 	      if (flags & POINTER)
 		*(va_arg (ap, _PTR *)) = (_PTR) (unsigned _POINTER_INT) res;
+	      else if (flags & CHAR)
+		{
+		  cp = va_arg (ap, char *);
+		  *cp = res;
+		}
 	      else if (flags & SHORT)
 		{
 		  sp = va_arg (ap, short *);
@@ -676,6 +951,18 @@ __svfscanf (fp, fmt0, ap)
 		  lp = va_arg (ap, long *);
 		  *lp = res;
 		}
+#ifndef _NO_LONGLONG
+	      else if (flags & LONGDBL)
+		{
+		  u_long_long resll;
+		  if (ccfn == _strtoul_r)
+		    resll = _strtoull_r (rptr, buf, (char **) NULL, base);
+		  else
+		    resll = _strtoll_r (rptr, buf, (char **) NULL, base);
+		  llp = va_arg (ap, long long*);
+		  *llp = resll;
+		}
+#endif
 	      else
 		{
 		  ip = va_arg (ap, int *);
@@ -683,9 +970,9 @@ __svfscanf (fp, fmt0, ap)
 		}
 	      nassigned++;
 	    }
-	  nread += p - buf;
+	  nread += p - buf + skips;
 	  break;
-
+	}
 #ifdef FLOATING_POINT
 	case CT_FLOAT:
 	{
@@ -697,15 +984,17 @@ __svfscanf (fp, fmt0, ap)
 	  long leading_zeroes = 0;
 	  long zeroes, exp_adjust;
 	  char *exp_start = NULL;
+	  unsigned width_left = 0;
 #ifdef hardway
 	  if (width == 0 || width > sizeof (buf) - 1)
-	    width = sizeof (buf) - 1;
 #else
 	  /* size_t is unsigned, hence this optimisation */
-	  if (--width > sizeof (buf) - 2)
-	    width = sizeof (buf) - 2;
-	  width++;
+	  if (width - 1 > sizeof (buf) - 2)
 #endif
+	    {
+	      width_left = width - (sizeof (buf) - 1);
+	      width = sizeof (buf) - 1;
+	    }
 	  flags |= SIGNOK | NDIGITS | DPTOK | EXPOK;
 	  zeroes = 0;
 	  exp_adjust = 0;
@@ -724,6 +1013,11 @@ __svfscanf (fp, fmt0, ap)
 		    {
 		      flags &= ~SIGNOK;
 		      zeroes++;
+		      if (width_left)
+			{
+			  width_left--;
+			  width++;
+			}
 		      goto fskip;
 		    }
 		  /* Fall through.  */
@@ -777,15 +1071,13 @@ __svfscanf (fp, fmt0, ap)
 	      break;
 	    fok:
 	      *p++ = c;
-	      width--;
 	    fskip:
+	      width--;
               ++nread;
 	      if (--fp->_r > 0)
 		fp->_p++;
 	      else
-#ifndef CYGNUS_NEC
 	      if (__srefill (fp))
-#endif
 		break;		/* EOF */
 	    }
 	  if (zeroes)
@@ -820,12 +1112,14 @@ __svfscanf (fp, fmt0, ap)
 	    }
 	  if ((flags & SUPPRESS) == 0)
 	    {
+	      double res = 0;
 #ifdef _NO_LONGDBL
-	      double res;
+#define QUAD_RES res;
 #else  /* !_NO_LONG_DBL */
-	      long double res;
+	      long double qres = 0;
+#define QUAD_RES qres;
 #endif /* !_NO_LONG_DBL */
-	      long new_exp;
+	      long new_exp = 0;
 
 	      *p = 0;
 	      if ((flags & (DPTOK | EXPOK)) == EXPOK)
@@ -835,7 +1129,7 @@ __svfscanf (fp, fmt0, ap)
 		  exp_start = p;
 		}
 	      else if (exp_adjust)
-		new_exp = atol (exp_start + 1) - exp_adjust;
+                new_exp = _strtol_r (rptr, (exp_start + 1), NULL, 10) - exp_adjust;
 	      if (exp_adjust)
 		{
 
@@ -845,11 +1139,17 @@ __svfscanf (fp, fmt0, ap)
 		    exp_start = buf + sizeof (buf) - MAX_LONG_LEN - 1;
                  sprintf (exp_start, "e%ld", new_exp);
 		}
-#ifdef _NO_LONGDBL
-	      res = atof (buf);
-#else  /* !_NO_LONGDBL */
-	      res = _strtold (buf, NULL);
-#endif /* !_NO_LONGDBL */
+
+	      /* Current _strtold routine is markedly slower than 
+	         _strtod_r.  Only use it if we have a long double
+	         result.  */
+#ifndef _NO_LONGDBL /* !_NO_LONGDBL */
+	      if (flags & LONGDBL)
+	      	qres = _strtold (buf, NULL);
+	      else
+#endif
+	        res = _strtod_r (rptr, buf, NULL);
+
 	      if (flags & LONG)
 		{
 		  dp = va_arg (ap, double *);
@@ -858,7 +1158,7 @@ __svfscanf (fp, fmt0, ap)
 	      else if (flags & LONGDBL)
 		{
 		  ldp = va_arg (ap, _LONG_DOUBLE *);
-		  *ldp = res;
+		  *ldp = QUAD_RES;
 		}
 	      else
 		{
@@ -873,111 +1173,10 @@ __svfscanf (fp, fmt0, ap)
 	}
     }
 input_failure:
+  _funlockfile (fp);
   return nassigned ? nassigned : -1;
 match_failure:
+all_done:
+  _funlockfile (fp);
   return nassigned;
 }
-
-/*
- * Fill in the given table from the scanset at the given format
- * (just after `[').  Return a pointer to the character past the
- * closing `]'.  The table has a 1 wherever characters should be
- * considered part of the scanset.
- */
-
-/*static*/
-u_char *
-__sccl (tab, fmt)
-     register char *tab;
-     register u_char *fmt;
-{
-  register int c, n, v;
-
-  /* first `clear' the whole table */
-  c = *fmt++;			/* first char hat => negated scanset */
-  if (c == '^')
-    {
-      v = 1;			/* default => accept */
-      c = *fmt++;		/* get new first char */
-    }
-  else
-    v = 0;			/* default => reject */
-  /* should probably use memset here */
-  for (n = 0; n < 256; n++)
-    tab[n] = v;
-  if (c == 0)
-    return fmt - 1;		/* format ended before closing ] */
-
-  /*
-   * Now set the entries corresponding to the actual scanset to the
-   * opposite of the above.
-   *
-   * The first character may be ']' (or '-') without being special; the
-   * last character may be '-'.
-   */
-
-  v = 1 - v;
-  for (;;)
-    {
-      tab[c] = v;		/* take character c */
-    doswitch:
-      n = *fmt++;		/* and examine the next */
-      switch (n)
-	{
-
-	case 0:		/* format ended too soon */
-	  return fmt - 1;
-
-	case '-':
-	  /*
-	   * A scanset of the form [01+-] is defined as `the digit 0, the
-	   * digit 1, the character +, the character -', but the effect of a
-	   * scanset such as [a-zA-Z0-9] is implementation defined.  The V7
-	   * Unix scanf treats `a-z' as `the letters a through z', but treats
-	   * `a-a' as `the letter a, the character -, and the letter a'.
-	   *
-	   * For compatibility, the `-' is not considerd to define a range if
-	   * the character following it is either a close bracket (required by
-	   * ANSI) or is not numerically greater than the character we just
-	   * stored in the table (c).
-	   */
-	  n = *fmt;
-	  if (n == ']' || n < c)
-	    {
-	      c = '-';
-	      break;		/* resume the for(;;) */
-	    }
-	  fmt++;
-	  do
-	    {			/* fill in the range */
-	      tab[++c] = v;
-	    }
-	  while (c < n);
-#if 1			/* XXX another disgusting compatibility hack */
-	  /*
-	   * Alas, the V7 Unix scanf also treats formats such
-	   * as [a-c-e] as `the letters a through e'. This too
-	   * is permitted by the standard....
-	   */
-	  goto doswitch;
-#else
-	  c = *fmt++;
-	  if (c == 0)
-	    return fmt - 1;
-	  if (c == ']')
-	    return fmt;
-#endif
-
-	  break;
-
-
-	case ']':		/* end of scanset */
-	  return fmt;
-
-	default:		/* just another character */
-	  c = n;
-	  break;
-	}
-    }
-  /* NOTREACHED */
-}
